/* Select-based JOH loop.
   Copyright (C) 2011 Sergey Poznyakoff

   JOH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   JOH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JOH.  If not, see <http://www.gnu.org/licenses/>. */

#include "joh.h"

static int
_netio_unused(int fd, struct joh_sockaddr *sa, void *data)
{
	die(EX_SOFTWARE, "internal error: unhandled descriptor %d",
	    fd);
}

static int
_netio_master(int fd, struct joh_sockaddr *sa, void *data)
{
	int sd;
	fd_set *fds = data;
	union {
		struct sockaddr_in s_in;
		struct sockaddr_in6 s_in6;
		struct sockaddr_un s_un;
		struct sockaddr s_sa;
	} s;
	socklen_t len = sizeof s;
	struct joh_sockaddr *jsa;
	
	if (FD_ISSET(fd, &fds[FDSET_EX])) {
		char c;
		recv(fd, &c, 1, MSG_OOB);
	}
	if (!FD_ISSET(fd, &fds[FDSET_RD]))
		return 0;
	sd = accept(fd, (struct sockaddr *) &s, &len);
	if (sd < 0) {
		error("accept(%s): %s",
		      joh_sockaddr_str(sa), strerror(errno));
		return 0; /* FIXME: perhaps 1? */
	}

	if (!connection_allowed(sd, sa->srvname)) {
		char *addr_string = sockaddr_str(&s.s_sa, len);
		error("access from %s blocked by TCP wrappers, service %s",
		      addr_string, sa->srvname);
		free(addr_string);
		close(sd);
		return 0;
	}

	jsa = joh_sockaddr_new(len);
	memcpy(jsa->addr, &s, len);
	joh_add_fd(sd, FDMASK(FDSET_RD));
	jsa->sess = joh_session_create(sd, conn_client);
	jsa->class = sa->class;
	jsa->srvname = sa->srvname;
	
	debug(20, ("created client session @ %s", joh_sockaddr_str(jsa)));
	joh_connection_register(sd, conn_client, jsa);
	
	return 0;
}

static int
_netio_jabber(int fd, struct joh_sockaddr *sa, void *data)
{
	fd_set *fds = data;
	struct joh_session *sess = sa->sess;
	ssize_t rc;
	
	if (FD_ISSET(fd, &fds[FDSET_EX])) {
		char c;
		if (recv(fd, &c, 1, MSG_OOB) < 1) {
			if (sess->peer)
				joh_session_destroy(sess->peer);
			joh_session_destroy(sess);
			return 0;
		}
	}

	if (FD_ISSET(fd, &fds[FDSET_WR]) &&
	    sess->obuf.off < sess->obuf.size) {
		rc = send(fd, sess->obuf.base + sess->obuf.off,
			  sess->obuf.size - sess->obuf.off, 0);
		if (rc == -1) {
			error("%s: write error on %s: %s",
			      conn_type_str[sess->type],
			      joh_sockaddr_str(sa), strerror(errno));
			joh_session_destroy(sess);
			return 0;
		}
		if (debug_level == 100) {
			char *pfx[4] = { "Jabber", NULL, "S", NULL };
			pfx[1] = (char*) joh_sockaddr_str(sa);
			joh_logdump(pfx, sess->obuf.base + sess->obuf.off, rc);
		}
		sess->obuf.off += rc;
		if (sess->obuf.off == sess->obuf.size) {
			joh_chg_fd(sess->fd, FDMASK(FDSET_RD));
			sess->obuf.base = NULL;
			sess->obuf.size = sess->obuf.off = 0;
			return 0;
		}
	}
	
	if (FD_ISSET(fd, &fds[FDSET_RD])) {
		char iobuf[IOBUFSIZE];
		
		rc = recv(fd, iobuf, sizeof(iobuf), 0);
		if (rc == -1) {
			error("%s: read error on %s: %s",
			      conn_type_str[sess->type],
			      joh_sockaddr_str(sa), strerror(errno));
			joh_session_destroy(sess);
			return 0;
		}
		if (debug_level == 100) {
			char *pfx[4] = { "Jabber", NULL, "C", NULL };
			pfx[1] = (char*) joh_sockaddr_str(sa);
			joh_logdump(pfx, iobuf, rc);
		}
		
		joh_cbuf_append(sess->ibuf, iobuf, rc);
		joh_chg_fd(sess->fd, FDMASK(FDSET_EX));
	}
	return 0;
}

static int
_netio_client(int fd, struct joh_sockaddr *sa, void *data)
{
	fd_set *fds = data;
	struct joh_session *sess = sa->sess;
	ssize_t rc;
	
	if (FD_ISSET(fd, &fds[FDSET_EX])) {
		char c;
		if (recv(fd, &c, 1, MSG_OOB) < 1) {
			joh_session_destroy(sess);
			return 0;
		}
	}

	if (FD_ISSET(fd, &fds[FDSET_RD])) {
		char iobuf[IOBUFSIZE];
		
		rc = recv(fd, iobuf, sizeof(iobuf), 0);
		if (rc == -1) {
			error("%s/%s: read error on %s: %s",
			      conn_type_str[sess->type],
			      client_class_str[sa->class],
			      joh_sockaddr_str(sa), strerror(errno));
			joh_session_destroy(sess);
			return 0;
		}
		if (debug_level == 100) {
			char *pfx[4] = { NULL, NULL, "C", NULL };
			pfx[0] = client_class_str[sa->class];
			pfx[1] = (char*) joh_sockaddr_str(sa);
			joh_logdump(pfx, iobuf, rc);
		}
		joh_cbuf_append(sess->ibuf, iobuf, rc);
		rc = joh_data_decode(sess, sa);
		if (rc == 0) {
			struct joh_session *js;

			joh_chg_fd(sess->fd, FDMASK(FDSET_WR));
			
			joh_cbuf_clear(sess->ibuf);
			
			js = joh_session_lookup(sess);
			if (!js) {
				strcpy(sess->param[joh_param_id].ptr, "0:0");
				sess->param[joh_param_id].size = 3;
				return 0;
			}

			if (!joh_validate_key(sess, js)) {
				error("%s/%s: connection from %s has invalid key",
				      conn_type_str[sess->type],
				      client_class_str[sa->class],
				      joh_sockaddr_str(sa));
				/* XEP-0025:
				   3.1.4 Key Sequence Error

				      Server returns ID=-3:0
				*/
				strcpy(sess->param[joh_param_id].ptr, "-3:0");
				sess->param[joh_param_id].size = 4;
				return 0;
			}				 

			if (sess->param[joh_param_newkey].ptr) {
				free(js->param[joh_param_key].ptr);
				js->param[joh_param_key].ptr =
					sess->param[joh_param_newkey].ptr;
				js->param[joh_param_key].size =
					sess->param[joh_param_newkey].size;
				sess->param[joh_param_newkey].ptr = NULL;
				sess->param[joh_param_newkey].size = 0;
			}
			
			sess->peer = js;
			js->peer = sess;
			if (sess->param[joh_param_payload].size) {
				joh_chg_fd(js->fd, FDMASK(FDSET_WR));
				js->obuf.base = sess->param[joh_param_payload].ptr;
				js->obuf.base_alloc = 0;
				js->obuf.size = sess->param[joh_param_payload].size;
				js->obuf.off = 0;
			} else
				joh_chg_fd(js->fd, FDMASK(FDSET_RD));
		} else if (rc < 0) {
			error("%s/%s: protocol error on %s: %s",
			      conn_type_str[sess->type],
			      client_class_str[sa->class],
			      joh_sockaddr_str(sa), strerror(errno));
			joh_session_destroy(sess);
			return 0;
		}
		return 0;
	}
	
	if (FD_ISSET(fd, &fds[FDSET_WR])) {
		if (!sess->obuf.base) {
			int rc;
			size_t size;
			static int param_names[2] = { joh_param_id,
						      joh_param_payload };
			struct joh_param param_vals[2];

			param_vals[0] = sess->param[joh_param_id];
			if (sess->peer &&
			    (size = joh_cbuf_size (sess->peer->ibuf))) {
				param_vals[1].ptr = joh_cbuf_finish(sess->peer->ibuf);
				param_vals[1].size = size;
				rc = joh_data_encode(sa,
						     &sess->obuf,
						     2, param_names,
						     param_vals);
				joh_cbuf_clear(sess->peer->ibuf);
				joh_chg_fd(sess->peer->fd, FDMASK(FDSET_RD));
			} else {
				rc = joh_data_encode(sa,
						     &sess->obuf,
						     1, param_names,
						     param_vals);
			}
			if (rc < 0) {
				joh_session_destroy(sess);
				return 0;
			}
		}
		rc = send(fd, sess->obuf.base + sess->obuf.off,
			  sess->obuf.size - sess->obuf.off, 0);
		if (rc == -1) {
			error("%s/%s: write error on %s: %s",
			      conn_type_str[sess->type],
			      client_class_str[sa->class],
			      joh_sockaddr_str(sa), strerror(errno));
			joh_session_destroy(sess);
		} else {
			if (debug_level == 100) {
				char *pfx[4] = { NULL, NULL, "S", NULL };
				pfx[0] = client_class_str[sa->class];
				pfx[1] = (char*) joh_sockaddr_str(sa);
				joh_logdump(pfx,
					    sess->obuf.base + sess->obuf.off,
					    rc);
			}
			sess->obuf.off += rc;
			if (sess->obuf.off == sess->obuf.size)
				joh_session_destroy(sess);
		}
	}

	return 0;
}

static int (*netio_handler[])(int, struct joh_sockaddr *, void *) = {
	_netio_unused,
	_netio_master,
	_netio_jabber,
	_netio_client
};

int
netio_itr(int fd, enum joh_conntype type, struct joh_sockaddr *sa, void *data)
{
	return netio_handler[type](fd, sa, data);
}

static int got_signal;

static RETSIGTYPE
sighan(int sig)
{
	got_signal = sig;
}

static void
set_signals(RETSIGTYPE (*handler)(int), int *sigv, int sigc)
{
	int i;
	struct sigaction act;

	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	for (i = 0; i < sigc; i++)
		sigaddset(&act.sa_mask, sigv[i]);

	for (i = 0; i < sigc; i++) {
		act.sa_handler = handler;
		sigaction(sigv[i], &act, NULL);
	}
}

int
joh_loop()
{
	int rc;
	fd_set fds[3];
	static int sigtab[] = { SIGPIPE, /*SIGABRT,*/ SIGINT, SIGQUIT,
				SIGTERM, SIGHUP };
	
	set_signals(sighan, sigtab, sizeof(sigtab)/sizeof(sigtab[0]));

	while (!got_signal) {
		struct timeval tv, *tvp;

		tvp = joh_session_timeout(&tv);
		memcpy(fds, joh_fd_set, sizeof(fds));

		rc = select(joh_max_fd + 1, &fds[0], &fds[1], &fds[2], tvp);
		if (rc == -1) {
			if (errno == EINTR)
				continue;
			die(EX_OSERR, "select: %s", strerror(errno));
		}
		joh_connection_iterate(netio_itr, fds);
	}
	set_signals(SIG_DFL, sigtab, sizeof(sigtab)/sizeof(sigtab[0]));
	return got_signal;
}
